@@ -20,31 +20,6 @@
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  *  Boston, MA 02110-1301 USA
-*/
-
-/**
- * SECTION:element-vaapidecode
- * @short_description: A VA-API based video decoder
- *
- * vaapidecode decodes from raw bitstreams to surfaces suitable for
- * the vaapisink or vaapipostproc elements using the installed <ulink
- * url="https://wiki.freedesktop.org/www/Software/vaapi/">VA-API</ulink>
- * back-end.
- *
- * In the case of OpenGL based elements, the buffers have the
- * #GstVideoGLTextureUploadMeta meta, which efficiently copies the
- * content of the VA-API surface into a GL texture.
- *
- * Also it can deliver normal video buffers that can be rendered or
- * processed by other elements, but the performance would be rather
- * bad.
- *
- * <refsect2>
- * <title>Example launch line</title>
- * |[
- * gst-launch-1.0 filesrc location=~/big_buck_bunny.mov ! qtdemux ! h264parse ! vaapidecode ! vaapisink
- * ]|
- * </refsect2>
  */
 
 #include "gstcompat.h"
@@ -104,8 +79,10 @@ static const char gst_vaapidecode_sink_caps_str[] =
 
 static const char gst_vaapidecode_src_caps_str[] =
     GST_VAAPI_MAKE_SURFACE_CAPS ";"
+#if (USE_GLX || USE_EGL)
     GST_VAAPI_MAKE_GLTEXUPLOAD_CAPS ";"
-    GST_VIDEO_CAPS_MAKE("{ I420, YV12, NV12 }");
+#endif
+    GST_VIDEO_CAPS_MAKE("{ NV12, I420, YV12, P010_10LE }");
 
 static GstStaticPadTemplate gst_vaapidecode_src_factory =
     GST_STATIC_PAD_TEMPLATE(
@@ -127,6 +104,23 @@ struct _GstVaapiDecoderMap
 static const GstVaapiDecoderMap vaapi_decode_map[] = {
 #if USE_JPEG_DECODER
   {GST_VAAPI_CODEC_JPEG, GST_RANK_MARGINAL, "jpeg", "image/jpeg"},
+#endif
+  {GST_VAAPI_CODEC_MPEG2, GST_RANK_PRIMARY, "mpeg2",
+      "video/mpeg, mpegversion=2, systemstream=(boolean)false"},
+  {GST_VAAPI_CODEC_MPEG4, GST_RANK_PRIMARY, "mpeg4",
+      "video/mpeg, mpegversion=4"},
+  {GST_VAAPI_CODEC_H263, GST_RANK_PRIMARY, "h263", "video/x-h263"},
+  {GST_VAAPI_CODEC_H264, GST_RANK_PRIMARY, "h264", "video/x-h264"},
+  {GST_VAAPI_CODEC_VC1, GST_RANK_PRIMARY, "vc1",
+      "video/x-wmv, wmvversion=3, format={WMV3,WVC1}"},
+#if USE_VP8_DECODER
+  {GST_VAAPI_CODEC_VP8, GST_RANK_PRIMARY, "vp8", "video/x-vp8"},
+#endif
+#if USE_VP9_DECODER
+  {GST_VAAPI_CODEC_VP9, GST_RANK_PRIMARY, "vp9", "video/x-vp9"},
+#endif
+#if USE_HEVC_DECODER
+  {GST_VAAPI_CODEC_H265, GST_RANK_PRIMARY, "h265", "video/x-h265"},
 #endif
   {0 /* the rest */ , GST_RANK_PRIMARY + 1, NULL,
       gst_vaapidecode_sink_caps_str},
@@ -139,8 +133,8 @@ static gboolean gst_vaapidecode_update_sink_caps (GstVaapiDecode * decode,
     GstCaps * caps);
 static gboolean gst_vaapi_decode_input_state_replace (GstVaapiDecode * decode,
     const GstVideoCodecState * new_state);
-static gboolean gst_vaapidecode_negotiate (GstVaapiDecode * decode);
 
+/* get invoked only if actural VASurface size (not the cropped values) changed */
 static void
 gst_vaapi_decoder_state_changed (GstVaapiDecoder * decoder,
     const GstVideoCodecState * codec_state, gpointer user_data)
@@ -153,8 +147,6 @@ gst_vaapi_decoder_state_changed (GstVaapiDecoder * decoder,
     return;
   if (!gst_vaapidecode_update_sink_caps (decode, decode->input_state->caps))
     return;
-
-  decode->do_renego = TRUE;
 }
 
 static GstVideoCodecState *
@@ -212,57 +204,105 @@ static gboolean
 gst_vaapidecode_update_src_caps (GstVaapiDecode * decode)
 {
   GstVideoDecoder *const vdec = GST_VIDEO_DECODER (decode);
+  GstPad *const srcpad = GST_VIDEO_DECODER_SRC_PAD (vdec);
+  GstCaps *templ;
   GstVideoCodecState *state, *ref_state;
   GstVaapiCapsFeature feature;
-  GstCapsFeatures *features = NULL;
+  GstCapsFeatures *features;
+  GstCaps *allocation_caps;
   GstVideoInfo *vi;
-  GstVideoFormat format = GST_VIDEO_FORMAT_I420;
+  GstVideoFormat format;
   GstClockTime latency;
   gint fps_d, fps_n;
+  guint width, height;
+  const gchar *format_str, *feature_str;
 
   if (!decode->input_state)
     return FALSE;
 
   ref_state = decode->input_state;
 
-  feature =
-      gst_vaapi_find_preferred_caps_feature (GST_VIDEO_DECODER_SRC_PAD (vdec),
-      GST_VIDEO_INFO_FORMAT (&ref_state->info), &format);
+  format = GST_VIDEO_INFO_FORMAT (&decode->decoded_info);
+  templ = gst_pad_get_pad_template_caps (srcpad);
+  feature = gst_vaapi_find_preferred_caps_feature (srcpad, templ, &format);
+  gst_caps_unref (templ);
 
   if (feature == GST_VAAPI_CAPS_FEATURE_NOT_NEGOTIATED)
     return FALSE;
 
-  switch (feature) {
-#if (USE_GLX || USE_EGL)
-    case GST_VAAPI_CAPS_FEATURE_GL_TEXTURE_UPLOAD_META:
-      features =
-          gst_caps_features_new
-          (GST_CAPS_FEATURE_META_GST_VIDEO_GL_TEXTURE_UPLOAD_META, NULL);
-      break;
+#if (!USE_GLX && !USE_EGL)
+  /* This is a very pathological situation. Should not happen. */
+  if (feature == GST_VAAPI_CAPS_FEATURE_GL_TEXTURE_UPLOAD_META)
+    return FALSE;
 #endif
-    case GST_VAAPI_CAPS_FEATURE_VAAPI_SURFACE:
-      features =
-          gst_caps_features_new (GST_CAPS_FEATURE_MEMORY_VAAPI_SURFACE, NULL);
-      break;
-    default:
-      break;
+
+  if ((feature == GST_VAAPI_CAPS_FEATURE_SYSTEM_MEMORY ||
+          feature == GST_VAAPI_CAPS_FEATURE_VAAPI_SURFACE)
+      && format != GST_VIDEO_INFO_FORMAT (&decode->decoded_info)) {
+    GST_FIXME_OBJECT (decode, "validate if driver can convert from %s to %s",
+        gst_video_format_to_string (GST_VIDEO_INFO_FORMAT
+            (&decode->decoded_info)), gst_video_format_to_string (format));
   }
 
-  state = gst_video_decoder_set_output_state (vdec, format,
-      ref_state->info.width, ref_state->info.height, ref_state);
-  if (!state || state->info.width == 0 || state->info.height == 0) {
-    if (features)
-      gst_caps_features_free (features);
-    if (state)
-      gst_video_codec_state_unref (state);
+  width = decode->display_width;
+  height = decode->display_height;
+
+  if (!width || !height) {
+    width = GST_VIDEO_INFO_WIDTH (&ref_state->info);
+    height = GST_VIDEO_INFO_HEIGHT (&ref_state->info);
+  }
+
+  state = gst_video_decoder_set_output_state (vdec, format, width, height,
+      ref_state);
+  if (!state)
+    return FALSE;
+
+  if (GST_VIDEO_INFO_WIDTH (&state->info) == 0
+      || GST_VIDEO_INFO_HEIGHT (&state->info) == 0) {
+    gst_video_codec_state_unref (state);
     return FALSE;
   }
 
   vi = &state->info;
-
   state->caps = gst_video_info_to_caps (vi);
-  if (features)
-    gst_caps_set_features (state->caps, 0, features);
+
+  switch (feature) {
+    case GST_VAAPI_CAPS_FEATURE_GL_TEXTURE_UPLOAD_META:
+    case GST_VAAPI_CAPS_FEATURE_VAAPI_SURFACE:{
+      GstStructure *structure = gst_caps_get_structure (state->caps, 0);
+
+      /* Remove chroma-site and colorimetry from src caps,
+       * which is unnecessary on downstream if using VASurface
+       */
+      gst_structure_remove_fields (structure, "chroma-site", "colorimetry",
+          NULL);
+
+      feature_str = gst_vaapi_caps_feature_to_string (feature);
+      features = gst_caps_features_new (feature_str, NULL);
+      gst_caps_set_features (state->caps, 0, features);
+      break;
+    }
+    default:
+      break;
+  }
+
+  /* Allocation query is different from pad's caps */
+  allocation_caps = NULL;
+  if (GST_VIDEO_INFO_WIDTH (&decode->decoded_info) != width
+      || GST_VIDEO_INFO_HEIGHT (&decode->decoded_info) != height) {
+    allocation_caps = gst_caps_copy (state->caps);
+    format_str = gst_video_format_to_string (format);
+    gst_caps_set_simple (allocation_caps,
+        "width", G_TYPE_INT, GST_VIDEO_INFO_WIDTH (&decode->decoded_info),
+        "height", G_TYPE_INT, GST_VIDEO_INFO_HEIGHT (&decode->decoded_info),
+        "format", G_TYPE_STRING, format_str, NULL);
+    GST_INFO_OBJECT (decode, "new alloc caps = %" GST_PTR_FORMAT,
+        allocation_caps);
+  }
+  gst_caps_replace (&state->allocation_caps, allocation_caps);
+  if (allocation_caps)
+    gst_caps_unref (allocation_caps);
+
   GST_INFO_OBJECT (decode, "new src caps = %" GST_PTR_FORMAT, state->caps);
   gst_caps_replace (&decode->srcpad_caps, state->caps);
   gst_video_codec_state_unref (state);
@@ -294,26 +334,97 @@ gst_vaapidecode_release (GstVaapiDecode * decode)
   gst_object_unref (decode);
 }
 
+/* check whether the decoded surface size has changed */
 static gboolean
-is_surface_resolution_changed (GstVideoDecoder * vdec,
+is_surface_resolution_changed (GstVaapiDecode * decode,
     GstVaapiSurface * surface)
 {
+  GstVideoInfo *vinfo = &decode->decoded_info;
+  GstVideoFormat surface_format;
   guint surface_width, surface_height;
-  guint configured_width, configured_height;
-  GstVideoCodecState *state;
-  gboolean ret = FALSE;
+
+  g_return_val_if_fail (surface != NULL, FALSE);
 
   gst_vaapi_surface_get_size (surface, &surface_width, &surface_height);
 
+  if (GST_VIDEO_INFO_WIDTH (vinfo) == surface_width
+      && GST_VIDEO_INFO_HEIGHT (vinfo) == surface_height)
+    return FALSE;
+
+  /* doing gst_vaapi_surface_get_format() only if necessary since it
+   * execute vaDeriveImage in the background. This will usually get
+   * executed only once */
+  surface_format = GST_VIDEO_INFO_FORMAT (vinfo);
+  if (surface_format == GST_VIDEO_FORMAT_UNKNOWN) {
+    surface_format = gst_vaapi_surface_get_format (surface);
+
+    /* if the VA context delivers a currently unrecognized format
+     * (ICM3, e.g.), we can assume NV12 "safely" */
+    if (surface_format == GST_VIDEO_FORMAT_UNKNOWN
+        || surface_format == GST_VIDEO_FORMAT_ENCODED)
+      surface_format = GST_VIDEO_FORMAT_NV12;
+  }
+
+  gst_video_info_set_format (vinfo, surface_format, surface_width,
+      surface_height);
+
+  return TRUE;
+}
+
+/* check whether display resolution changed */
+static gboolean
+is_display_resolution_changed (GstVaapiDecode * decode,
+    const GstVaapiRectangle * crop_rect)
+{
+  GstVideoDecoder *const vdec = GST_VIDEO_DECODER (decode);
+  GstVideoCodecState *state;
+  guint display_width, display_height;
+  guint negotiated_width, negotiated_height;
+
+  display_width = GST_VIDEO_INFO_WIDTH (&decode->decoded_info);
+  display_height = GST_VIDEO_INFO_HEIGHT (&decode->decoded_info);
+  if (crop_rect) {
+    display_width = crop_rect->width;
+    display_height = crop_rect->height;
+  }
+
   state = gst_video_decoder_get_output_state (vdec);
-  configured_width = GST_VIDEO_INFO_WIDTH (&state->info);
-  configured_height = GST_VIDEO_INFO_HEIGHT (&state->info);
+  if (G_UNLIKELY (!state))
+    goto set_display_res;
+
+  negotiated_width = GST_VIDEO_INFO_WIDTH (&state->info);
+  negotiated_height = GST_VIDEO_INFO_HEIGHT (&state->info);
   gst_video_codec_state_unref (state);
 
-  if (surface_width != configured_width || surface_height != configured_height)
-    ret = TRUE;
+  if ((display_width == negotiated_width && display_height == negotiated_height)
+      && (decode->display_width == negotiated_width
+          && decode->display_height == negotiated_height))
+    return FALSE;
 
-  return ret;
+set_display_res:
+  decode->display_width = display_width;
+  decode->display_height = display_height;
+  return TRUE;
+}
+
+static gboolean
+gst_vaapidecode_negotiate (GstVaapiDecode * decode)
+{
+  GstVideoDecoder *const vdec = GST_VIDEO_DECODER (decode);
+  GstVaapiPluginBase *const plugin = GST_VAAPI_PLUGIN_BASE (vdec);
+
+  GST_DEBUG_OBJECT (decode, "Input codec state changed, doing renegotiation");
+
+  if (!gst_vaapi_plugin_base_set_caps (plugin, decode->sinkpad_caps, NULL))
+    return FALSE;
+  if (!gst_vaapidecode_update_src_caps (decode))
+    return FALSE;
+  if (!gst_video_decoder_negotiate (vdec))
+    return FALSE;
+  if (!gst_vaapi_plugin_base_set_caps (plugin, NULL, decode->srcpad_caps))
+    return FALSE;
+
+  return TRUE;
 }
 
 static GstFlowReturn
@@ -322,18 +433,36 @@ gst_vaapidecode_push_decoded_frame (GstVideoDecoder * vdec,
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
   GstVaapiSurfaceProxy *proxy;
+  GstVaapiSurface *surface;
   GstFlowReturn ret;
   const GstVaapiRectangle *crop_rect;
   GstVaapiVideoMeta *meta;
   guint flags, out_flags = 0;
+  gboolean alloc_renegotiate, caps_renegotiate;
 
   if (!GST_VIDEO_CODEC_FRAME_IS_DECODE_ONLY (out_frame)) {
     proxy = gst_video_codec_frame_get_user_data (out_frame);
+    surface = GST_VAAPI_SURFACE_PROXY_SURFACE (proxy);
+    crop_rect = gst_vaapi_surface_proxy_get_crop_rect (proxy);
 
-    /* reconfigure if un-cropped surface resolution changed */
-    if (is_surface_resolution_changed (vdec,
-            GST_VAAPI_SURFACE_PROXY_SURFACE (proxy)))
-      gst_vaapidecode_negotiate (decode);
+    /* in theory, we are not supposed to check the surface resolution
+     * change here since it should be advertised before from ligstvaapi.
+     * But there are issues with it especially for some vp9 streams where
+     * upstream element set un-cropped values in set_format() which make
+     * everything a mess. So better doing the explicit check here irrespective
+     * of what notification we get from upstream or libgstvaapi.Also, even if
+     * we received notification from libgstvaapi, the frame we are going to
+     * be pushed at this point might not have the notified resolution if there
+     * are queued frames in decoded picture buffer. */
+    alloc_renegotiate = is_surface_resolution_changed (decode, surface);
+    caps_renegotiate = is_display_resolution_changed (decode, crop_rect);
+
+    if (gst_pad_needs_reconfigure (GST_VIDEO_DECODER_SRC_PAD (vdec))
+        || alloc_renegotiate || caps_renegotiate) {
+
+      if (!gst_vaapidecode_negotiate (decode))
+        return GST_FLOW_ERROR;
+    }
 
     gst_vaapi_surface_proxy_set_destroy_notify (proxy,
         (GDestroyNotify) gst_vaapidecode_release, gst_object_ref (decode));
@@ -366,7 +495,6 @@ gst_vaapidecode_push_decoded_frame (GstVideoDecoder * vdec,
           GST_VIDEO_BUFFER_FLAG_FIRST_IN_BUNDLE);
     }
 
-    crop_rect = gst_vaapi_surface_proxy_get_crop_rect (proxy);
     if (crop_rect) {
       GstVideoCropMeta *const crop_meta =
           gst_buffer_add_video_crop_meta (out_frame->output_buffer);
@@ -383,11 +511,16 @@ gst_vaapidecode_push_decoded_frame (GstVideoDecoder * vdec,
 #endif
   }
 
+  if (decode->in_segment.rate < 0.0
+      && !GST_VIDEO_CODEC_FRAME_IS_SYNC_POINT (out_frame)) {
+    GST_TRACE_OBJECT (decode, "drop frame in reverse playback");
+    gst_video_decoder_release_frame (GST_VIDEO_DECODER (decode), out_frame);
+    return GST_FLOW_OK;
+  }
+
   ret = gst_video_decoder_finish_frame (vdec, out_frame);
   if (ret != GST_FLOW_OK)
     goto error_commit_buffer;
-
-  gst_video_codec_frame_unref (out_frame);
   return GST_FLOW_OK;
 
   /* ERRORS */
@@ -401,7 +534,6 @@ error_create_buffer:
         ("video sink failed to create video buffer for proxy'ed "
             "surface %" GST_VAAPI_ID_FORMAT, GST_VAAPI_ID_ARGS (surface_id)));
     gst_video_decoder_drop_frame (vdec, out_frame);
-    gst_video_codec_frame_unref (out_frame);
     return GST_FLOW_ERROR;
   }
 error_get_meta:
@@ -410,43 +542,16 @@ error_get_meta:
         ("Failed to get vaapi video meta attached to video buffer"),
         ("Failed to get vaapi video meta attached to video buffer"));
     gst_video_decoder_drop_frame (vdec, out_frame);
-    gst_video_codec_frame_unref (out_frame);
     return GST_FLOW_ERROR;
   }
 error_commit_buffer:
   {
     GST_INFO_OBJECT (decode, "downstream element rejected the frame (%s [%d])",
         gst_flow_get_name (ret), ret);
-    gst_video_codec_frame_unref (out_frame);
     return ret;
   }
 }
 
-static gboolean
-gst_vaapidecode_negotiate (GstVaapiDecode * decode)
-{
-  GstVideoDecoder *const vdec = GST_VIDEO_DECODER (decode);
-  GstVaapiPluginBase *const plugin = GST_VAAPI_PLUGIN_BASE (vdec);
-
-  if (!decode->do_renego)
-    return TRUE;
-
-  GST_DEBUG_OBJECT (decode, "Input codec state changed, doing renegotiation");
-
-  if (!gst_vaapi_plugin_base_set_caps (plugin, decode->sinkpad_caps, NULL))
-    return FALSE;
-  if (!gst_vaapidecode_update_src_caps (decode))
-    return FALSE;
-  if (!gst_video_decoder_negotiate (vdec))
-    return FALSE;
-  if (!gst_vaapi_plugin_base_set_caps (plugin, NULL, decode->srcpad_caps))
-    return FALSE;
-
-  decode->do_renego = FALSE;
-
-  return TRUE;
-}
-
 static GstFlowReturn
 gst_vaapidecode_push_all_decoded_frames (GstVaapiDecode * decode)
 {
@@ -460,17 +565,13 @@ gst_vaapidecode_push_all_decoded_frames (GstVaapiDecode * decode)
 
     switch (status) {
       case GST_VAAPI_DECODER_STATUS_SUCCESS:
+        /* GstVaapiDecode's queue adds an extra reference */
+        gst_video_codec_frame_unref (out_frame);
         ret = gst_vaapidecode_push_decoded_frame (vdec, out_frame);
         if (ret != GST_FLOW_OK)
           return ret;
         break;
       case GST_VAAPI_DECODER_STATUS_ERROR_NO_DATA:
-        /* Delayed the pool re-negotiation untill we push all decoded (and queued)
-         * frames downstream. Otherwise for the multi-resolution videos, the
-         * GstVideoVideoMemory will be having wrong resolution.
-         * commit 6eba201f3252eba6a99ab7da7a4c662091a3e884 */
-        if (!gst_vaapidecode_negotiate (decode))
-          return GST_FLOW_ERROR;
         return GST_FLOW_OK;
       default:
         GST_VIDEO_DECODER_ERROR (vdec, 1, STREAM, DECODE, ("Decoding failed"),
@@ -487,28 +588,11 @@ gst_vaapidecode_handle_frame (GstVideoDecoder * vdec,
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
   GstVaapiDecoderStatus status;
-  GstVaapiPluginBase *plugin;
   GstFlowReturn ret;
 
   if (!decode->input_state)
     goto not_negotiated;
 
-  if (G_UNLIKELY (!decode->active) ||
-      gst_pad_needs_reconfigure (GST_VIDEO_DECODER_SRC_PAD (vdec))) {
-    GST_DEBUG_OBJECT (decode, "activating the decoder");
-    if (!gst_vaapidecode_update_src_caps (decode))
-      goto not_negotiated;
-
-    if (!gst_video_decoder_negotiate (vdec))
-      goto not_negotiated;
-
-    plugin = GST_VAAPI_PLUGIN_BASE (vdec);
-    if (!gst_vaapi_plugin_base_set_caps (plugin, NULL, decode->srcpad_caps))
-      goto not_negotiated;
-
-    decode->active = TRUE;
-  }
-
   /* Decode current frame */
   for (;;) {
     status = gst_vaapi_decoder_decode (decode->decoder, frame);
@@ -569,6 +653,17 @@ not_negotiated:
   }
 }
 
+/* If there is something in GstVideoDecoder's output adapter, then
+   submit the frame for decoding */
+static inline void
+gst_vaapidecode_flush_output_adapter (GstVaapiDecode * decode)
+{
+  if (decode->current_frame_size == 0)
+    return;
+  gst_video_decoder_have_frame (GST_VIDEO_DECODER (decode));
+  decode->current_frame_size = 0;
+}
+
 static GstFlowReturn
 gst_vaapidecode_drain (GstVideoDecoder * vdec)
 {
@@ -577,49 +672,35 @@ gst_vaapidecode_drain (GstVideoDecoder * vdec)
   if (!decode->decoder)
     return GST_FLOW_NOT_NEGOTIATED;
 
+  GST_LOG_OBJECT (decode, "drain");
+
+  gst_vaapidecode_flush_output_adapter (decode);
   return gst_vaapidecode_push_all_decoded_frames (decode);
 }
 
-static gboolean
-gst_vaapidecode_internal_flush (GstVideoDecoder * vdec)
-{
-  GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
-  GstVaapiDecoderStatus status;
-
-  if (!decode->decoder)
-    return TRUE;
-
-  /* If there is something in GstVideoDecoder's output adapter, then
-     submit the frame for decoding */
-  if (decode->current_frame_size) {
-    gst_video_decoder_have_frame (vdec);
-    decode->current_frame_size = 0;
-  }
-
-  status = gst_vaapi_decoder_flush (decode->decoder);
-  if (status != GST_VAAPI_DECODER_STATUS_SUCCESS) {
-    GST_WARNING_OBJECT (decode, "failed to flush decoder (status %d)", status);
-    return FALSE;
-  }
-
-  return TRUE;
-}
-
 static GstFlowReturn
 gst_vaapidecode_finish (GstVideoDecoder * vdec)
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
-  gboolean flushed;
+  GstVaapiDecoderStatus status;
   GstFlowReturn ret;
 
   if (!decode->decoder)
     return GST_FLOW_OK;
 
-  flushed = gst_vaapidecode_internal_flush (vdec);
+  gst_vaapidecode_flush_output_adapter (decode);
+  status = gst_vaapi_decoder_flush (decode->decoder);
   ret = gst_vaapidecode_push_all_decoded_frames (decode);
-  if (!flushed)
-    return GST_FLOW_ERROR;
+  if (status != GST_VAAPI_DECODER_STATUS_SUCCESS)
+    goto error_decoder_flush;
   return ret;
+
+  /* ERRORS: */
+error_decoder_flush:
+  {
+    GST_WARNING_OBJECT (decode, "failed to flush decoder (status %d)", status);
+    return GST_FLOW_ERROR;
+  }
 }
 
 static gboolean
@@ -629,7 +710,11 @@ gst_vaapidecode_decide_allocation (GstVideoDecoder * vdec, GstQuery * query)
   GstCaps *caps = NULL;
 
   gst_query_parse_allocation (query, &caps, NULL);
+  if (!caps)
+    goto error_no_caps;
+
   decode->has_texture_upload_meta = FALSE;
+
 #if (USE_GLX || USE_EGL)
   decode->has_texture_upload_meta =
       gst_query_find_allocation_meta (query,
@@ -639,7 +724,14 @@ gst_vaapidecode_decide_allocation (GstVideoDecoder * vdec, GstQuery * query)
 #endif
 
   return gst_vaapi_plugin_base_decide_allocation (GST_VAAPI_PLUGIN_BASE (vdec),
-      query, 0);
+      query);
+
+  /* ERRORS */
+error_no_caps:
+  {
+    GST_ERROR_OBJECT (decode, "no caps specified");
+    return FALSE;
+  }
 }
 
 static inline gboolean
@@ -783,8 +875,6 @@ gst_vaapidecode_destroy (GstVaapiDecode * decode)
   gst_vaapi_decoder_replace (&decode->decoder, NULL);
   gst_caps_replace (&decode->decoder_caps, NULL);
 
-  decode->active = FALSE;
-
   gst_vaapidecode_release (gst_object_ref (decode));
 }
 
@@ -814,9 +904,7 @@ gst_vaapidecode_finalize (GObject * object)
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (object);
 
-  gst_caps_replace (&decode->sinkpad_caps, NULL);
-  gst_caps_replace (&decode->srcpad_caps, NULL);
-  gst_caps_replace (&decode->allowed_caps, NULL);
+  gst_caps_replace (&decode->allowed_sinkpad_caps, NULL);
 
   g_cond_clear (&decode->surface_ready);
   g_mutex_clear (&decode->surface_ready_mutex);
@@ -829,22 +917,15 @@ static gboolean
 gst_vaapidecode_open (GstVideoDecoder * vdec)
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
-  GstVaapiDisplay *const old_display = GST_VAAPI_PLUGIN_BASE_DISPLAY (decode);
-  gboolean success;
 
   if (!gst_vaapi_plugin_base_open (GST_VAAPI_PLUGIN_BASE (decode)))
     return FALSE;
 
-  /* Let GstVideoContext ask for a proper display to its neighbours */
-  /* Note: steal old display that may be allocated from get_caps()
-     so that to retain a reference to it, thus avoiding extra
-     initialization steps if we turn out to simply re-use the
-     existing (cached) VA display */
-  GST_VAAPI_PLUGIN_BASE_DISPLAY (decode) = NULL;
-  success = gst_vaapidecode_ensure_display (decode);
-  if (old_display)
-    gst_vaapi_display_unref (old_display);
-  return success;
+  decode->display_width = 0;
+  decode->display_height = 0;
+  gst_video_info_init (&decode->decoded_info);
+
+  return TRUE;
 }
 
 static gboolean
@@ -852,23 +933,66 @@ gst_vaapidecode_close (GstVideoDecoder * vdec)
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
 
-  gst_vaapi_decode_input_state_replace (decode, NULL);
   gst_vaapidecode_destroy (decode);
   gst_vaapi_plugin_base_close (GST_VAAPI_PLUGIN_BASE (decode));
   return TRUE;
 }
 
+static gboolean
+gst_vaapidecode_start (GstVideoDecoder * vdec)
+{
+  GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
+  GstVaapiDisplay *const old_display = GST_VAAPI_PLUGIN_BASE_DISPLAY (decode);
+  gboolean success;
+
+  /* Let GstVideoContext ask for a proper display to its neighbours */
+  /* Note: steal old display that may be allocated from get_caps()
+     so that to retain a reference to it, thus avoiding extra
+     initialization steps if we turn out to simply re-use the
+     existing (cached) VA display */
+  GST_VAAPI_PLUGIN_BASE_DISPLAY (decode) = NULL;
+  success =
+      gst_vaapi_plugin_base_ensure_display (GST_VAAPI_PLUGIN_BASE (decode));
+  if (old_display)
+    gst_vaapi_display_unref (old_display);
+
+  return success;
+}
+
+static gboolean
+gst_vaapidecode_stop (GstVideoDecoder * vdec)
+{
+  GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
+
+  gst_vaapidecode_purge (decode);
+  gst_vaapi_decode_input_state_replace (decode, NULL);
+  gst_vaapi_decoder_replace (&decode->decoder, NULL);
+  gst_caps_replace (&decode->decoder_caps, NULL);
+  gst_caps_replace (&decode->sinkpad_caps, NULL);
+  gst_caps_replace (&decode->srcpad_caps, NULL);
+  return TRUE;
+}
+
 static gboolean
 gst_vaapidecode_flush (GstVideoDecoder * vdec)
 {
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
+  gboolean reverse;
 
-  if (decode->decoder && !gst_vaapidecode_internal_flush (vdec))
+  if (!decode->decoder)
     return FALSE;
 
+  GST_LOG_OBJECT (vdec, "flushing");
+
+  gst_vaapidecode_purge (decode);
+
+  /* in reverse playback we cannot destroy the decoder at flush, since
+   * it will lost the parsing state */
+  reverse = decode->in_segment.rate < 0;
+
   /* There could be issues if we avoid the reset_full() while doing
    * seeking: we have to reset the internal state */
-  return gst_vaapidecode_reset_full (decode, decode->sinkpad_caps, TRUE);
+  return gst_vaapidecode_reset_full (decode, decode->sinkpad_caps, !reverse);
 }
 
 static gboolean
@@ -883,7 +1007,7 @@ gst_vaapidecode_set_format (GstVideoDecoder * vdec, GstVideoCodecState * state)
     return FALSE;
   if (!gst_vaapi_plugin_base_set_caps (plugin, decode->sinkpad_caps, NULL))
     return FALSE;
-  if (!gst_vaapidecode_reset_full (decode, decode->sinkpad_caps, FALSE))
+  if (!gst_vaapidecode_reset_full (decode, decode->sinkpad_caps, TRUE))
     return FALSE;
 
   return TRUE;
@@ -946,9 +1070,9 @@ gst_vaapidecode_parse (GstVideoDecoder * vdec,
 }
 
 static gboolean
-gst_vaapidecode_ensure_allowed_caps (GstVaapiDecode * decode)
+gst_vaapidecode_ensure_allowed_sinkpad_caps (GstVaapiDecode * decode)
 {
-  GstCaps *caps, *allowed_caps;
+  GstCaps *caps, *allowed_sinkpad_caps;
   GArray *profiles;
   guint i;
 
@@ -958,8 +1082,8 @@ gst_vaapidecode_ensure_allowed_caps (GstVaapiDecode * decode)
   if (!profiles)
     goto error_no_profiles;
 
-  allowed_caps = gst_caps_new_empty ();
-  if (!allowed_caps)
+  allowed_sinkpad_caps = gst_caps_new_empty ();
+  if (!allowed_sinkpad_caps)
     goto error_no_memory;
 
   for (i = 0; i < profiles->len; i++) {
@@ -983,9 +1107,9 @@ gst_vaapidecode_ensure_allowed_caps (GstVaapiDecode * decode)
       gst_structure_set (structure, "profile", G_TYPE_STRING,
           profile_name, NULL);
 
-    allowed_caps = gst_caps_merge (allowed_caps, caps);
+    allowed_sinkpad_caps = gst_caps_merge (allowed_sinkpad_caps, caps);
   }
-  decode->allowed_caps = gst_caps_simplify (allowed_caps);
+  decode->allowed_sinkpad_caps = gst_caps_simplify (allowed_sinkpad_caps);
 
   g_array_unref (profiles);
   return TRUE;
@@ -1010,7 +1134,7 @@ gst_vaapidecode_sink_getcaps (GstVideoDecoder * vdec, GstCaps * filter)
   GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
   GstCaps *result;
 
-  if (decode->allowed_caps)
+  if (decode->allowed_sinkpad_caps)
     goto bail;
 
   /* if we haven't a display yet, return our pad's template caps */
@@ -1019,11 +1143,12 @@ gst_vaapidecode_sink_getcaps (GstVideoDecoder * vdec, GstCaps * filter)
 
   /* if the allowed caps calculation fails, return an empty caps, so
    * the auto-plug can try other decoder */
-  if (!gst_vaapidecode_ensure_allowed_caps (decode))
+  if (!gst_vaapidecode_ensure_allowed_sinkpad_caps (decode))
     return gst_caps_new_empty ();
 
 bail:
-  result = gst_video_decoder_proxy_getcaps (vdec, decode->allowed_caps, filter);
+  result = gst_video_decoder_proxy_getcaps (vdec, decode->allowed_sinkpad_caps,
+      filter);
 
   GST_DEBUG_OBJECT (decode, "Returning sink caps %" GST_PTR_FORMAT, result);
 
@@ -1089,6 +1214,27 @@ gst_vaapidecode_src_query (GstVideoDecoder * vdec, GstQuery * query)
   return ret;
 }
 
+static gboolean
+gst_vaapidecode_sink_event (GstVideoDecoder * vdec, GstEvent * event)
+{
+  GstVaapiDecode *const decode = GST_VAAPIDECODE (vdec);
+
+  switch (GST_EVENT_TYPE (event)) {
+    case GST_EVENT_SEGMENT:
+    {
+      /* Keep segment event to refer to rate so that
+       * vaapidecode can handle reverse playback
+       */
+      gst_event_copy_segment (event, &decode->in_segment);
+      break;
+    }
+    default:
+      break;
+  }
+
+  return GST_VIDEO_DECODER_CLASS (parent_class)->sink_event (vdec, event);
+}
+
 static void
 gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
 {
@@ -1097,7 +1243,7 @@ gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
   GstVideoDecoderClass *const vdec_class = GST_VIDEO_DECODER_CLASS (klass);
   GstPadTemplate *pad_template;
   GstVaapiDecoderMap *map;
-  gchar *name, *longname;
+  gchar *name, *longname, *description;
   GstCaps *caps;
 
   GST_DEBUG_CATEGORY_INIT (gst_debug_vaapidecode,
@@ -1111,6 +1257,8 @@ gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
 
   vdec_class->open = GST_DEBUG_FUNCPTR (gst_vaapidecode_open);
   vdec_class->close = GST_DEBUG_FUNCPTR (gst_vaapidecode_close);
+  vdec_class->start = GST_DEBUG_FUNCPTR (gst_vaapidecode_start);
+  vdec_class->stop = GST_DEBUG_FUNCPTR (gst_vaapidecode_stop);
   vdec_class->set_format = GST_DEBUG_FUNCPTR (gst_vaapidecode_set_format);
   vdec_class->flush = GST_DEBUG_FUNCPTR (gst_vaapidecode_flush);
   vdec_class->parse = GST_DEBUG_FUNCPTR (gst_vaapidecode_parse);
@@ -1122,6 +1270,7 @@ gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
   vdec_class->src_query = GST_DEBUG_FUNCPTR (gst_vaapidecode_src_query);
   vdec_class->sink_query = GST_DEBUG_FUNCPTR (gst_vaapidecode_sink_query);
   vdec_class->getcaps = GST_DEBUG_FUNCPTR (gst_vaapidecode_sink_getcaps);
+  vdec_class->sink_event = GST_DEBUG_FUNCPTR (gst_vaapidecode_sink_event);
 
   map = (GstVaapiDecoderMap *) g_type_get_qdata (G_OBJECT_CLASS_TYPE (klass),
       GST_VAAPI_DECODE_PARAMS_QDATA);
@@ -1129,20 +1278,23 @@ gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
   if (map->codec) {
     name = g_ascii_strup (map->name, -1);
     longname = g_strdup_printf ("VA-API %s decoder", name);
+    description = g_strdup_printf ("A VA-API based %s video decoder", name);
     g_free (name);
   } else {
     longname = g_strdup ("VA-API decoder");
+    description = g_strdup (GST_PLUGIN_DESC);
   }
 
   element_class->set_context = gst_vaapi_base_set_context;
   gst_element_class_set_static_metadata (element_class, longname,
-      "Codec/Decoder/Video", GST_PLUGIN_DESC,
+      "Codec/Decoder/Video", description,
       "Gwenole Beauchesne <gwenole.beauchesne@intel.com>, "
       "Halley Zhao <halley.zhao@intel.com>, "
       "Sreerenj Balachandran <sreerenj.balachandran@intel.com>, "
       "Wind Yuan <feng.yuan@intel.com>");
 
   g_free (longname);
+  g_free (description);
 
   /* sink pad */
   caps = gst_caps_from_string (map->caps_str);
@@ -1152,8 +1304,8 @@ gst_vaapidecode_class_init (GstVaapiDecodeClass * klass)
   gst_element_class_add_pad_template (element_class, pad_template);
 
   /* src pad */
-  pad_template = gst_static_pad_template_get (&gst_vaapidecode_src_factory);
-  gst_element_class_add_pad_template (element_class, pad_template);
+  gst_element_class_add_static_pad_template (element_class,
+      &gst_vaapidecode_src_factory);
 }
 
 static void
@@ -1163,10 +1315,6 @@ gst_vaapidecode_init (GstVaapiDecode * decode)
 
   gst_vaapi_plugin_base_init (GST_VAAPI_PLUGIN_BASE (decode), GST_CAT_DEFAULT);
 
-  decode->decoder = NULL;
-  decode->decoder_caps = NULL;
-  decode->allowed_caps = NULL;
-
   g_mutex_init (&decode->surface_ready_mutex);
   g_cond_init (&decode->surface_ready);
 
@@ -1174,7 +1322,7 @@ gst_vaapidecode_init (GstVaapiDecode * decode)
 }
 
 gboolean
-gst_vaapidecode_register (GstPlugin * plugin)
+gst_vaapidecode_register (GstPlugin * plugin, GArray * decoders)
 {
   gboolean ret = FALSE;
   guint i, codec, rank;
@@ -1198,6 +1346,9 @@ gst_vaapidecode_register (GstPlugin * plugin)
     rank = vaapi_decode_map[i].rank;
     name = vaapi_decode_map[i].name;
 
+    if (codec && !gst_vaapi_codecs_has_codec (decoders, codec))
+      continue;
+
     if (codec) {
       type_name = g_strdup_printf ("GstVaapiDecode_%s", name);
       element_name = g_strdup_printf ("vaapi%sdec", name);
@@ -1216,7 +1367,11 @@ gst_vaapidecode_register (GstPlugin * plugin)
           (gpointer) & vaapi_decode_map[i]);
     }
 
-    ret |= gst_element_register (plugin, element_name, rank, type);
+    /* Register GstVaapiDecode as GObject type, but not in GStreamer, so
+     * vaapidecodebin can use it internally, but no exposed as a plugin
+     * feature */
+    if (codec)
+      ret |= gst_element_register (plugin, element_name, rank, type);
 
     g_free (element_name);
     g_free (type_name);
