@@ -41,17 +41,20 @@ struct _GstVaapiDecode {
 
     GstCaps            *sinkpad_caps;
     GstCaps            *srcpad_caps;
+    GstVideoInfo        decoded_info;
     GstVaapiDecoder    *decoder;
     GMutex              surface_ready_mutex;
     GCond               surface_ready;
     GstCaps            *decoder_caps;
-    GstCaps            *allowed_caps;
+    GstCaps            *allowed_sinkpad_caps;
     guint               current_frame_size;
     guint               has_texture_upload_meta : 1;
 
+    guint               display_width;
+    guint               display_height;
+
     GstVideoCodecState *input_state;
-    volatile gboolean   active;
-    volatile gboolean   do_renego;
+    GstSegment          in_segment;
 };
 
 struct _GstVaapiDecodeClass {
@@ -59,7 +62,7 @@ struct _GstVaapiDecodeClass {
     GstVaapiPluginBaseClass parent_class;
 };
 
-gboolean gst_vaapidecode_register (GstPlugin * plugin);
+gboolean gst_vaapidecode_register (GstPlugin * plugin, GArray * decoders);
 
 G_END_DECLS
 
